<?php

namespace Snaju\Inception\Http\Request\Validators;

use Snaju\Inception\Http\Request\Validators\Rules\ObjectRule;
use Snaju\Inception\Http\Request\Validators\Rules\BooleanRule;
use Snaju\Inception\Http\Request\Validators\Rules\GreaterThanEqualRule;
use Snaju\Inception\Http\Request\Validators\Rules\GreaterThanRule;
use Snaju\Inception\Http\Request\Validators\Rules\InArrayRule;
use Snaju\Inception\Http\Request\Validators\Rules\IntegerRule;
use Snaju\Inception\Http\Request\Validators\Rules\JsonRule;
use Snaju\Inception\Http\Request\Validators\Rules\LessThan;
use Snaju\Inception\Http\Request\Validators\Rules\LessThanEqual;
use Snaju\Inception\Http\Request\Validators\Rules\MaxRule;
use Snaju\Inception\Http\Request\Validators\Rules\MinRule;
use Snaju\Inception\Http\Request\Validators\Rules\NumericRule;
use Snaju\Inception\Http\Request\Validators\Rules\PresentRule;
use Snaju\Inception\Http\Request\Validators\Rules\RequiredRule;
use Snaju\Inception\Http\Request\Validators\Rules\StringRule;
use Snaju\Inception\Exception\InvalidArgumentException;
use Snaju\Inception\Exception\InvalidKeyException;
use Snaju\Inception\Http\Request\WebRequest;
use Snaju\Inception\InceptionCore;
use Snaju\Inception\Util\Loaders\Collection;


class Validator implements ValidatorInterface
{

    private WebRequest $request;

    private array $rules = [];

    private array $data = [];

    private array $errors = [];

    private Collection $defaultRules;

    private bool $usesRequestData = true;

//    private array $allowedParentObjects = ['headers', 'cookies', 'query', 'body'];

    public function __construct(array $rules = [], array $data = [], bool $sendErrorResponses = true)
    {
        $this->request = InceptionCore::getRequest();
        $this->defaultRules = new Collection([
            'array'         => ObjectRule::class,
            'bool'          => BooleanRule::class,
            'gte'           => GreaterThanEqualRule::class,
            'gt'            => GreaterThanRule::class,
            'in_array'      => InArrayRule::class,
            'integer'       => IntegerRule::class,
            'json'          => JsonRule::class,
            'lt'            => LessThan::class,
            'lte'           => LessThanEqual::class,
            'max'           => MaxRule::class,
            'min'           => MinRule::class,
            'numeric'       => NumericRule::class,
//            'object'        => ObjectRule::class,
            'present'       => PresentRule::class, // should be first in validator string, bails if fail
            'required'      => RequiredRule::class, // should be first in validator string, bails if fail
            'string'        => StringRule::class,
        ]);

        $this->rules = $rules;

        // Use the request data if no data was specified
        if (count($data) > 0) {
            $this->data = $data;
            $this->usesRequestData = false;
        }
            else {
                $this->data = $this->request->__dangerouslytoArray();
                $this->usesRequestData = true;
            }

        if (count($rules) > 0) {
            $this->validate($this->data, $sendErrorResponses);
        }
    }

    public function validate(array $data = [], bool $sendErrorResponses = false)
    {

//        $ruleArray = $this->expandDotSyntaxArrayKeys($this->rules);

        $ruleArray = array_merge([], $this->rules);

//        echo '<pre>';
//        print_r($ruleArray);
//        echo '</pre>';



//        return response()->json([
////            'rules' => $fieldRules,
//            'data' => $data,
//            'ruleArray' => $ruleArray,
//            'sendErrorResponses' => $sendErrorResponses,
//        ]);


        foreach ($ruleArray as $key => $value) {
            if (str_contains($key, '.')) {
                return;
//                $nestedArray = $this->
//               echo $key;
            }
                else {
                    $rules = $this->createRulesObjectFromString($value);


//                    print_r($rules);
//                    echo (bool)property_exists($rules, 'if_present');
//                    exit;
                    $optional = property_exists($rules, 'if_present');

//                    echo '<pre>';
//                    echo 'these are the rules below' . "\n";
//                    print_r($rules);
//                    echo '</pre>';

//                    if (in_array($key, $data)) {
//                        print_r($rules);



                        foreach ($rules as $ruleId => $ruleValue) {
//                            $optional = false;

//                            if ($ruleId === 'if_present') {
//                                $optional = true;
//                            }

                            if ($this->defaultRules->keyExists($ruleId)) {
                                // Retrieve the static class of the validator
                                $namespace = $this->defaultRules->get($ruleId);
                                $validator = new $namespace($key, $ruleValue, $this->usesRequestData, $optional);


//                                echo $key;

//                                echo 'echo value: ' . $ruleId . ' ' . $ruleValue . '<br />';


                                $keyIsValid = $validator->isValid($data);
                                
                                if (! $keyIsValid) {
                                    $this->addError($key, $validator->getErrorMessage());

                                    if ($validator->stopOnError()) {
                                        break;
                                    }
                                }


//                                print_r($validator);
                            }
                                else if (! in_array($ruleId, ['if_present'])) {
                                    // TODO: throw an exception as say no validator with that key exists
                                    throw new InvalidKeyException("No validator with the the key `$ruleId` exists.");
                                }

                        }
//                    }
                }
//

//            if (in_array($key, $data)) {
//
//            }
        }












//        $this->addError('name', 'The name field is required');
//        $this->addError('name', 'Name must be at least 2 characters.');



//        return response()->json($ruleArray);


        // Grab all the data



        // Validate

        // If an id doesn't have any dot syntax, just look at the data as is.
        // But if using request data, look into the body of the request.

        // If id has dot syntax, just look at the data as is












//        exit;


        if ($this->hasErrors() && $sendErrorResponses) {
            $this->sendErrorResponse();
        }
    }

    public function hasErrors(): bool
    {
        $errors = false;

        if (count($this->errors) > 0) {
            $errors = true;
        }

        return $errors;
    }

    public function getErrors(): array
    {
        return $this->errors;
    }

    public function addRule(string $identifier, $rule)
    {
        if ($rule instanceof RuleInterface) {
            $this->defaultRules->add($identifier, $rule);
        }
        else {
            throw new InvalidArgumentException('The validation Rule must implement the RuleInterface');
        }
    }

    private function addError(string $key, string $error)
    {
        $this->errors[$key][] = $error;
    }

    private function sendErrorResponse()
    {
        // TODO: in the future take into consideration the Accept to check what type of response to send instead of just JSON responses
        return response()->json(['errors' => $this->getErrors()], 400);
    }

    private function expandDotSyntaxArrayKeys(array $arr): array
    {
        $target = [];

        foreach($arr as $key => $value) {
            if (is_array($value)) {
                $value = $this->expandDotSyntaxArrayKeys($value);
            }

//            $value = $this->createRulesObjectFromString($value);

            foreach(array_reverse(explode('.', $key)) as $key) {
                $value = [$key => $value];
            }

            $target = array_merge_recursive($target, $value);
        }

        return $target;
    }

    private function createRulesObjectFromString(string $rulesString): object
    {
        $target = [];
        $stringRules = explode('|', $rulesString);

        foreach ($stringRules as $stringRule) {
            $separator = explode(':', $stringRule);
            $rule = $separator[0];
            $value = isset($separator[1]) ? $separator[1] : 'true';

             // If the value contains comma separated values, make an array
//            if (str_contains($value, ',')) {
//                $value = explode(',', $value);
//            }

            $target[$rule] = $value;
        }

        return (object)$target;
    }

}


// sequence / rules / steps
/***
 * If using data directly from request check all and no dot syntax 
 */

